ALL_TYPES = {'AllTypes': {'extensibility-implied': False,
              'imports': {},
              'types': {'Bitstring': {'size': None, 'type': 'BIT STRING'},
                        'Bitstring2': {'size': [9], 'type': 'BIT STRING'},
                        'Bitstring3': {'size': [(5, 7)], 'type': 'BIT STRING'},
                        'Bmpstring': {'type': 'BMPString'},
                        'Boolean': {'type': 'BOOLEAN'},
                        'Enumerated': {'type': 'ENUMERATED',
                                       'values': {1: 'one'}},
                        'Ia5string': {'type': 'IA5String'},
                        'Ia5string2': {'type': 'IA5String'},
                        'Ia5string3': {'type': 'IA5String'},
                        'Ia5string4': {'type': 'IA5String'},
                        'Integer': {'type': 'INTEGER'},
                        'Integer2': {'restricted-to': [(1, 99)],
                                     'type': 'INTEGER'},
                        'Integer3': {'restricted-to': [(1, 2)],
                                     'type': 'INTEGER'},
                        'Null': {'type': 'NULL'},
                        'Numericstring': {'type': 'NumericString'},
                        'Objectidentifier': {'type': 'OBJECT IDENTIFIER'},
                        'Octetstring': {'size': None, 'type': 'OCTET STRING'},
                        'Octetstring2': {'size': [2], 'type': 'OCTET STRING'},
                        'Octetstring3': {'size': [3], 'type': 'OCTET STRING'},
                        'Octetstring4': {'size': [(3, 7)],
                                         'type': 'OCTET STRING'},
                        'Octetstring5': {'size': [4], 'type': 'OCTET STRING'},
                        'Printablestring': {'type': 'PrintableString'},
                        'Real': {'type': 'REAL'},
                        'Real2': {'type': 'REAL'},
                        'Real3': {'type': 'REAL'},
                        'Real4': {'type': 'REAL'},
                        'Real5': {'type': 'REAL'},
                        'Real6': {'type': 'REAL'},
                        'Sequence': {'members': [], 'type': 'SEQUENCE'},
                        'Sequence2': {'members': [{'default': 0,
                                                   'name': 'a',
                                                   'optional': False,
                                                   'type': 'INTEGER'}],
                                      'type': 'SEQUENCE'},
                        'SequenceOf': {'element': {'type': 'INTEGER'},
                                       'size': None,
                                       'type': 'SEQUENCE OF'},
                        'Set': {'members': [], 'type': 'SET'},
                        'Set2': {'members': [{'default': 1,
                                              'name': 'a',
                                              'optional': False,
                                              'type': 'INTEGER'}],
                                 'type': 'SET'},
                        'SetOf': {'element': {'type': 'INTEGER'},
                                  'size': None,
                                  'type': 'SET OF'},
                        'Teletexstring': {'type': 'TeletexString'},
                        'Universalstring': {'type': 'UniversalString'},
                        'Utctime': {'type': 'UTCTime'},
                        'Utf8string': {'type': 'UTF8String'},
                        'Visiblestring': {'type': 'VisibleString'}},
              'values': {'bitstring': {'type': 'BIT',
                                       'value': ["'10001000 1001'B"]},
                         'bitstring2': {'type': 'BIT',
                                        'value': ["'01234567 89ABCD EF'H"]},
                         'boolean': {'type': 'BOOLEAN', 'value': ['TRUE']},
                         'boolean2': {'type': 'BOOLEAN', 'value': ['FALSE']},
                         'choice': {'type': 'CHOICE',
                                    'value': ['foo', ':', ['bar']]},
                         'ia5string': {'type': 'IA5String',
                                       'value': ['{', '1', ',', '2', '}']},
                         'ia5string2': {'type': 'IA5String',
                                        'value': ['{',
                                                  '1',
                                                  ',',
                                                  '2',
                                                  ',',
                                                  '3',
                                                  ',',
                                                  '4',
                                                  '}']}}}}



FOO = {
    'Foo': {
        'extensibility-implied': False,
        'imports': {},
        'types': {
            'Answer': {
                'type': 'SEQUENCE',
                'members': [
                    {'name': 'id', 'optional': False, 'type': 'INTEGER'},
                    {'name': 'answer','optional': False,'type': 'BOOLEAN'}
                ]
            },
            'Question': {
                'type': 'SEQUENCE',
                'members': [
                    {'name': 'id', 'optional': False, 'type': 'INTEGER'},
                    {'name': 'question', 'optional': False, 'type': 'IA5String'}
                ]
            }
        },
        'values': {}
    }
}



"""Convert ASN.1 specifications to Python data structures.

"""

import logging

from pyparsing import Literal
from pyparsing import Keyword
from pyparsing import Word
from pyparsing import ZeroOrMore
from pyparsing import Regex
from pyparsing import printables
from pyparsing import delimitedList
from pyparsing import Group
from pyparsing import Optional
from pyparsing import Forward
from pyparsing import StringEnd
from pyparsing import OneOrMore
from pyparsing import alphanums
from pyparsing import nums
from pyparsing import Suppress
from pyparsing import ParseException
from pyparsing import ParseSyntaxException
from pyparsing import NotAny
from pyparsing import NoMatch


LOGGER = logging.getLogger(__name__)


class ParseError(Exception):
    pass


def convert_number(token):
    try:
        return int(token)
    except ValueError:
        return token


def convert_size(tokens):
    if len(tokens) == 0:
        return None
    elif '..' in tokens:
        return [(convert_number(tokens[0]),
                 convert_number(tokens[2]))]
    else:
        return [int(tokens[0])]


def convert_enum_values(tokens):
    number = 0
    values = {}

    for token in tokens:
        if len(token) == 2:
            number = int(token[1])

        values[number] = token[0]
        number += 1

    return values


def convert_tag(tokens):
    if len(tokens) > 0:
        if len(tokens[0]) == 1:
            tag = {
                'number': int(tokens[0][0])
            }
        else:
            tag = {
                'number': int(tokens[0][1]),
                'class': tokens[0][0]
            }

        if tokens[1]:
            tag['kind'] = tokens[1][0] if tokens[1] else None

        return tag


def convert_members(tokens):
    members = []

    for member_tokens in tokens:
        if member_tokens == ['...']:
            member_tokens = [['...', [], ''], []]

        member_tokens, qualifiers = member_tokens
        member = convert_type(member_tokens[2:])
        member['name'] = member_tokens[0]
        member['optional'] = 'OPTIONAL' in qualifiers

        if 'DEFAULT' in qualifiers:
            member['default'] = convert_number(qualifiers[1])

        tag = convert_tag(member_tokens[1])

        if tag:
            member['tag'] = tag

        members.append(member)

    return members


def convert_type(tokens):
    if tokens[0:2] == ['SEQUENCE', '{']:
        converted_type = {
            'type': 'SEQUENCE',
            'members': convert_members(tokens[2])
        }
    elif tokens[0] == 'SEQUENCE' and tokens[2] == 'OF':
        converted_type = {
            'type': 'SEQUENCE OF',
            'element': convert_type(tokens[4:]),
            'size': convert_size(tokens[1][2:-1])
        }

        tag = convert_tag(tokens[3])

        if tag:
            converted_type['element']['tag'] = tag
    elif tokens[0:2] == ['SET', '{']:
        converted_type = {
            'type': 'SET',
            'members': convert_members(tokens[2])
        }
    elif tokens[0] == 'SET' and tokens[2] == 'OF':
        converted_type = {
            'type': 'SET OF',
            'element': convert_type(tokens[4:]),
            'size': convert_size(tokens[1][2:-1])
        }

        tag = convert_tag(tokens[3])

        if tag:
            converted_type['element']['tag'] = tag
    elif tokens[0:2] == ['CHOICE', '{']:
        converted_type = {
            'type': 'CHOICE',
            'members': convert_members(tokens[2])
        }
    elif tokens[0] == 'INTEGER':
        converted_type = {'type': 'INTEGER'}

        if '..' in tokens[1]:
            minimum = convert_number(tokens[1][1])
            maximum = convert_number(tokens[1][3])
            converted_type['restricted-to'] = [(minimum, maximum)]
    elif tokens[0] == 'BOOLEAN':
        converted_type = {'type': 'BOOLEAN'}
    elif tokens[0:2] == ['ENUMERATED', '{']:
        converted_type = {
            'type': 'ENUMERATED',
            'values': convert_enum_values(tokens[2])
        }
    elif tokens[0:2] == ['OBJECT', 'IDENTIFIER']:
        converted_type = {'type': 'OBJECT IDENTIFIER'}
    elif tokens[0:2] == ['BIT', 'STRING']:
        converted_type = {'type': 'BIT STRING',
                          'size': convert_size(tokens[3][2:-1])}
    elif tokens[0:2] == ['OCTET', 'STRING']:
        converted_type = {'type': 'OCTET STRING',
                          'size': convert_size(tokens[2][2:-1])}
    elif tokens[0] == 'IA5String':
        converted_type = {'type': 'IA5String'}
    elif tokens[0:3] == ['ANY', 'DEFINED', 'BY']:
        converted_type = {
            'type': 'ANY DEFINED BY',
            'value': tokens[3],
            'choices': {}
        }
    else:
        converted_type = {'type': tokens[0]}

    return converted_type


def convert_type_tokens(tokens):
    converted_type = convert_type(tokens[3:])
    tag = convert_tag(tokens[2])

    if tag:
        converted_type['tag'] = tag

    return converted_type


def convert_value_tokens(tokens):
    type_ = tokens[1]

    if type_ == ['INTEGER']:
        value_type = 'INTEGER'
        value = int(tokens[3][0])
    elif type_ == ['OBJECT', 'IDENTIFIER']:
        value_type = 'OBJECT IDENTIFIER'
        value = []

        for value_tokens in tokens[3]:
            if len(value_tokens) == 2:
                value.append((value_tokens[0], int(value_tokens[1])))
            else:
                value.append(convert_number(value_tokens[0]))
    else:
        value_type = type_[0]
        value = tokens[3]

    return {'type': value_type, 'value': value}


def create_grammar():
    '''Return the ASN.1 grammar as Pyparsing objects.

    '''

    # Keywords.
    SEQUENCE = Keyword('SEQUENCE')
    CHOICE = Keyword('CHOICE')
    ENUMERATED = Keyword('ENUMERATED')
    DEFINITIONS = Keyword('DEFINITIONS')
    BEGIN = Keyword('BEGIN')
    END = Keyword('END')
    AUTOMATIC = Keyword('AUTOMATIC')
    TAGS = Keyword('TAGS')
    OPTIONAL = Keyword('OPTIONAL')
    OF = Keyword('OF')
    SIZE = Keyword('SIZE')
    INTEGER = Keyword('INTEGER')
    REAL = Keyword('REAL')
    BIT = Keyword('BIT')
    STRING = Keyword('STRING')
    OCTET = Keyword('OCTET')
    DEFAULT = Keyword('DEFAULT')
    IMPORTS = Keyword('IMPORTS')
    FROM = Keyword('FROM')
    CONTAINING = Keyword('CONTAINING')
    IMPLICIT = Keyword('IMPLICIT')
    EXPLICIT = Keyword('EXPLICIT')
    OBJECT = Keyword('OBJECT')
    IDENTIFIER = Keyword('IDENTIFIER')
    APPLICATION = Keyword('APPLICATION')
    PRIVATE = Keyword('PRIVATE')
    SET = Keyword('SET')
    ANY = Keyword('ANY')
    DEFINED = Keyword('DEFINED')
    BY = Keyword('BY')
    EXTENSIBILITY = Keyword('EXTENSIBILITY')
    IMPLIED = Keyword('IMPLIED')
    BOOLEAN = Keyword('BOOLEAN')
    TRUE = Keyword('TRUE')
    FALSE = Keyword('FALSE')

    # Various literals.
    word = Word(printables, excludeChars=',(){}[].:=;"|')
    type_reference = (NotAny(END) + Regex(r'[A-Z][a-zA-Z0-9-]*'))
    identifier = Regex(r'[a-z][a-zA-Z0-9-]*')
    value_reference = identifier
    value_name = Word(alphanums + '-')
    assign = Literal('::=')
    lparen = Literal('(')
    rparen = Literal(')')
    lbrace = Literal('{')
    rbrace = Literal('}')
    lbracket = Literal('[')
    rbracket = Literal(']')
    colon = Literal(':')
    scolon = Literal(';')
    dot = Literal('.')
    dotx2 = Literal('..')
    dotx3 = Literal('...')
    qmark = Literal('"')
    pipe = Literal('|')
    comma = Literal(',')
    integer = Word(nums)
    real_number = Regex(r'[+-]?\d+\.?\d*([eE][+-]?\d+)?')
    bstring = Regex(r"'[01\s]*'B")
    hstring = Regex(r"'[0-9A-F\s]*'H")
    cstring = NoMatch()
    number = word

    # Forward declarations.
    sequence_type = Forward().setName('SEQUENCE')
    choice_type = Forward().setName('CHOICE')
    integer_type = Forward().setName('INTEGER')
    real_type = Forward().setName('REAL')
    bit_string_type = Forward().setName('BIT STRING')
    octet_string_type = Forward().setName('OCTET STRING')
    enumerated_type = Forward().setName('ENUMERATED')
    sequence_of_type = Forward().setName('SEQUENCE OF')
    set_of_type = Forward().setName('SET OF')
    set_type = Forward().setName('SET')
    object_identifier_type = Forward().setName('OBJECT IDENTIFIER')
    boolean_type = Forward().setName('BOOLEAN')
    any_defined_by_type = Forward().setName('ANY DEFINED BY')

    bit_string_value = Forward()
    boolean_value = Forward()
    character_string_value = Forward()
    choice_value = Forward()

    range_ = (word + dotx2 + word)

    size = (SIZE
            + lparen
            + delimitedList(range_ | word, delim=pipe)
            + rparen)

    size_paren = (Suppress(Optional(lparen))
                  + size
                  + Suppress(Optional(rparen)))

    unions = Suppress(FROM
                      + delimitedList(qmark + word + qmark
                                      + dotx2
                                      + qmark + word + qmark, delim=pipe))

    element_set_spec = unions

    root_element_set_spec = element_set_spec

    root_element_set_specs = root_element_set_spec

    element_set_specs = root_element_set_specs

    subtype_constraint = element_set_specs

    constraint_spec = (size | subtype_constraint)

    constraint = (Suppress(lparen)
                  + constraint_spec
                  + Suppress(rparen))

    builtin_type = (choice_type
                    | integer_type
                    | real_type
                    | bit_string_type
                    | octet_string_type
                    | enumerated_type
                    | sequence_of_type
                    | sequence_type
                    | set_of_type
                    | set_type
                    | object_identifier_type
                    | boolean_type)

    type_ = (builtin_type
             | any_defined_by_type
             | (word + Optional(constraint)).setName('...'))

    tag = Group(Optional(Suppress(lbracket)
                         + Group(Optional(APPLICATION | PRIVATE) + word)
                         + Suppress(rbracket)
                         + Group(Optional(IMPLICIT | EXPLICIT))))

    sequence_type <<= (SEQUENCE
                       - lbrace
                       + Group(Optional(delimitedList(
                           Group(Group(identifier
                                       - tag
                                       - type_)
                                 + Group(Optional(OPTIONAL)
                                         + Optional(DEFAULT + word))
                                 | dotx3))))
                       - rbrace)

    sequence_of_type <<= (SEQUENCE
                          + Group(Optional(size_paren))
                          + OF
                          - tag
                          - type_)

    set_of_type <<= (SET
                     + Group(Optional(size))
                     + OF
                     - tag
                     - type_)

    set_type <<= (SET
                  - lbrace
                  + Group(Optional(delimitedList(
                      Group(Group(identifier
                                  - tag
                                  - type_)
                            + Group(Optional(OPTIONAL)
                                    + Optional(DEFAULT + word))
                            | dotx3))))
                  - rbrace)

    choice_type <<= (CHOICE
                     - lbrace
                     + Group(Optional(delimitedList(
                         Group(Group(identifier
                                     - tag
                                     - type_)
                               + Group(Optional(OPTIONAL)
                                       + Optional(DEFAULT + word))
                               | dotx3))))
                     - rbrace)

    enumerated_type <<= (ENUMERATED
                         - lbrace
                         + Group(delimitedList(Group((word
                                                      + Optional(Suppress(lparen)
                                                                 + word
                                                                 + Suppress(rparen)))
                                                     | dotx3)))
                         - rbrace)

    integer_type <<= (INTEGER
                      + Group(Optional((lparen
                                        + delimitedList(range_ | word, delim=pipe)
                                        + rparen)
                                       | (lbrace
                                          + Group(delimitedList(word
                                                                + lparen
                                                                + word
                                                                + rparen))
                                          + rbrace)))
                      + Optional(lparen
                                 + range_
                                 + rparen))

    real_type <<= (REAL
                   + Optional(lparen
                              + ((integer + dot + dotx2)
                                 | (integer + dotx2)
                                 | (real_number + dotx2))
                              + real_number
                              + rparen))

    bit_string_type <<= (BIT - STRING
                         + Group(Optional((lbrace
                                           + Group(delimitedList(word
                                                                 + lparen
                                                                 + word
                                                                 + rparen))
                                           + rbrace)))
                         + Group(Optional(constraint)))

    octet_string_type <<= (OCTET - STRING
                           + Group(Optional(Suppress(lparen)
                                            + (size | (CONTAINING + word))
                                            + Suppress(rparen))))

    object_identifier_type <<= (OBJECT - IDENTIFIER
                                + Optional(lparen
                                           + delimitedList(word, delim='|')
                                           + rparen))

    boolean_type <<= BOOLEAN

    any_defined_by_type <<= (ANY + DEFINED + BY + word)

    type_assignment = (type_reference
                       - assign
                       - tag
                       - type_)

    oid = (Suppress(lbrace)
           + ZeroOrMore(Group((value_name
                               + Suppress(lparen)
                               + word
                               + Suppress(rparen))
                              | word))
           + Suppress(rbrace))

    identifier_list = delimitedList(identifier)

    builtin_value = (bit_string_value
                     | boolean_value
                     | character_string_value
                     | choice_value
                     | word)

    value = Group(oid | builtin_value)

    bit_string_value <<= (bstring
                          | hstring
                          | (lbrace + Optional(identifier_list) + rbrace)
                          | (CONTAINING - value))

    boolean_value <<= (TRUE | FALSE)

    charsyms = NoMatch()

    character_string_list = (lbrace + charsyms + rbrace)

    group = number
    plane = number
    row = number
    cell = number
    quadruple = (lbrace
                 + group + comma
                 + plane + comma
                 + row + comma +
                 cell
                 + rbrace)

    table_column = number
    table_row = number
    tuple_ = (lbrace + table_column + comma + table_row + rbrace)

    restricted_character_string_value = (cstring
                                         | character_string_list
                                         | quadruple
                                         | tuple_)

    unrestricted_character_string_value = NoMatch()

    character_string_value <<= (restricted_character_string_value
                                | unrestricted_character_string_value)

    choice_value <<= (identifier + colon + value)

    value_assignment = (value_reference
                        - Group(INTEGER
                                | type_)
                        - assign
                        - value)

    assignment = Group(type_assignment
                       | value_assignment)

    symbols_imported = (Group(delimitedList(word))
                        + FROM
                        + word
                        + Suppress(Optional(oid)))

    imports = Group(Optional(IMPORTS
                             + symbols_imported
                             + scolon))

    assignment_list = Group(ZeroOrMore(assignment))

    module_body = (imports + assignment_list)

    module_reference = word

    definitive_identification = Group(Optional(oid))

    module_identifier = (module_reference + definitive_identification)

    tag_default = Group(Optional((AUTOMATIC | EXPLICIT | IMPLICIT) + TAGS))

    extension_default = Group(Optional(EXTENSIBILITY + IMPLIED))

    module_definition = Group(Group(module_identifier
                                    - DEFINITIONS
                                    + tag_default
                                    + extension_default
                                    - assign
                                    - BEGIN)
                              + module_body
                              - END)

    # The whole specification.
    specification = OneOrMore(module_definition) + StringEnd()
    comment = (Regex(r"--[\s\S]*?(--|\n)") | Regex(r"--(?:\\\n|[^\n])*"))
    specification.ignore(comment)

    return specification


def parse_string(string):
    """Parse given ASN.1 specification string and return a dictionary of
    its contents.

    >>> with open('foo.asn') as fin:
    ...     foo = asn1tools.parse_string(fin.read())

    """

    grammar = create_grammar()

    try:
        tokens = grammar.parseString(string).asList()
    except ParseException as e:
        raise ParseError("Invalid ASN.1 syntax at line {}, column {}: '{}'.".format(
            e.lineno,
            e.column,
            e.markInputline()))
    except ParseSyntaxException as e:
        raise ParseError("Invalid ASN.1 syntax at line {}, column {}: '{}': {}.".format(
            e.lineno,
            e.column,
            e.markInputline(),
            e.msg))

    modules = {}

    for module in tokens:
        module_name = module[0][0]

        LOGGER.debug("Found module '%s'.", module_name)

        imports = {}
        types = {}
        values = {}

        if module[1]:
            from_name = module[1][-2]
            LOGGER.debug("Found imports from '%s'.", from_name)
            imports[from_name] = module[1][1]

        for definition in module[2]:
            name = definition[0]

            if name[0].isupper():
                LOGGER.debug("Found type '%s'.", name)
                types[name] = convert_type_tokens(definition)
            else:
                LOGGER.debug("Found value '%s'.", name)
                values[name] = convert_value_tokens(definition)

        modules[module_name] = {
            'imports': imports,
            'types': types,
            'values': values
        }

        if module[0][3]:
            modules[module_name]['tags'] = module[0][3][0]

        modules[module_name]['extensibility-implied'] = (module[0][4] != [])

    return modules


def parse_files(filenames):
    """Parse given ASN.1 specification file(s) and return a dictionary of
    its/their contents.

    >>> foo = asn1tools.parse_files('foo.asn')

    """

    if isinstance(filenames, str):
        filenames = [filenames]

    string = ''

    for filename in filenames:
        with open(filename, 'r') as fin:
            string += fin.read()
            string += '\n'

    return parse_string(string)



"""The top level of the ASN.1 tools package contains commonly used
functions and classes, and the command line interface.

"""

import sys
import argparse
import binascii

from .compiler import compile_dict, compile_string, compile_files
from .parser import parse_string, parse_files, ParseError
from .codecs import EncodeError, DecodeError
from .errors import CompileError


__author__ = 'Erik Moqvist'
__version__ = '0.16.0'


def print_list(list_, indent):
    for i, element in enumerate(list_):
        print('{}[{}]:'.format(indent * ' ', i))
        if isinstance(element, list):
            print_list(element, indent + 2)
        elif isinstance(element, dict):
            print_dict(element, indent + 2)
        elif isinstance(element, tuple):
            decoded = binascii.hexlify(element[0]).decode('ascii')
            print('{}"{}"'.format(indent * ' ', decoded))
        elif isinstance(element, bytearray):
            decoded = binascii.hexlify(element).decode('ascii')
            print("{}'{}'".format(indent * ' ', decoded))
        else:
            print('{}{}'.format(indent * ' ', element))


def print_dict(dict_, indent=0):
    for key, value in dict_.items():
        if isinstance(value, list):
            print('{}{}:'.format(indent * ' ', key))
            print_list(value, indent + 2)
        elif isinstance(value, dict):
            print('{}{}:'.format(indent * ' ', key))
            print_dict(value, indent + 2)
        elif isinstance(value, tuple):
            decoded = binascii.hexlify(value[0]).decode('ascii')
            print('{}{}: "{}"'.format(indent * ' ', key, decoded))
        elif isinstance(value, bytearray):
            decoded = binascii.hexlify(value).decode('ascii')
            print("{}{}: '{}'".format(indent * ' ', key, decoded))
        else:
            print('{}{}: {}'.format(indent * ' ', key, value))


def _do_decode(args):
    specification = compile_files(args.specification, args.codec)
    encoded = binascii.unhexlify(args.hexstring)
    print_dict(specification.decode(args.type, encoded))


def _main():
    parser = argparse.ArgumentParser(
        description='Various ASN.1 utilities.')

    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('--version',
                        action='version',
                        version=__version__,
                        help='Print version information and exit.')

    # Workaround to make the subparser required in Python 3.
    subparsers = parser.add_subparsers(title='subcommands',
                                       dest='subcommand')
    subparsers.required = True

    # The 'decode' subparser.
    decode_parser = subparsers.add_parser(
        'decode',
        description='Decode given hextring and print it to standard output.')
    decode_parser.add_argument('-c', '--codec',
                               choices=('ber', 'der', 'jer', 'per', 'uper', 'xer'),
                               default='ber',
                               help='Codec (default: ber).')
    decode_parser.add_argument('specification', help='ASN.1 specification file (.asn).')
    decode_parser.add_argument('type', help='Type to decode.')
    decode_parser.add_argument('hexstring', help='Hexstring to decode.')
    decode_parser.set_defaults(func=_do_decode)

    args = parser.parse_args()

    if args.debug:
        args.func(args)
    else:
        try:
            args.func(args)
        except BaseException as e:
            sys.exit(str(e))



import sys
import unittest

import asn1tools

sys.path.append('tests/files')

from foo import FOO
from rrc_8_6_0 import RRC_8_6_0
from rfc5280 import RFC5280
from zforce import ZFORCE
from bar import BAR
from all_types import ALL_TYPES


class Asn1ToolsParseTest(unittest.TestCase):

    maxDiff = None

    def test_parse_foo(self):
        foo = asn1tools.parse_files('tests/files/foo.asn')
        self.assertEqual(foo, FOO)

    def test_parse_rrc_8_6_0(self):
        rrc_8_6_0 = asn1tools.parse_files('tests/files/rrc_8_6_0.asn')
        self.assertEqual(rrc_8_6_0, RRC_8_6_0)

    def test_parse_rfc5280(self):
        rfc5280 = asn1tools.parse_files('tests/files/rfc5280.asn')
        self.assertEqual(rfc5280, RFC5280)

    def test_parse_zforce(self):
        zforce = asn1tools.parse_files('tests/files/zforce.asn')
        self.assertEqual(zforce, ZFORCE)

    def test_parse_bar(self):
        bar = asn1tools.parse_files('tests/files/bar.asn')
        self.assertEqual(bar, BAR)

    def test_parse_all_types(self):
        all_types = asn1tools.parse_files('tests/files/all_types.asn')
        self.assertEqual(all_types, ALL_TYPES)

    def test_parse_error_empty_string(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('')

        self.assertEqual(str(cm.exception),
                         "Invalid ASN.1 syntax at line 1, column 1: '>!<'.")

    def test_parse_error_begin_missing(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= END')

        self.assertEqual(str(cm.exception),
                         "Invalid ASN.1 syntax at line 1, column 19: "
                         "'A DEFINITIONS ::= >!<END': Expected \"BEGIN\".")

    def test_parse_error_end_missing(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN')

        self.assertEqual(str(cm.exception),
                         "Invalid ASN.1 syntax at line 1, column 24: "
                         "'A DEFINITIONS ::= BEGIN>!<': Expected \"END\".")

    def test_parse_error_type_assignment_missing_assignment(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN A END')

        self.assertEqual(str(cm.exception),
                         "Invalid ASN.1 syntax at line 1, column 27: "
                         "'A DEFINITIONS ::= BEGIN A >!<END': "
                         "Expected \"::=\".")

    def test_parse_error_value_assignment_missing_assignment(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN a INTEGER END')

        self.assertEqual(str(cm.exception),
                         "Invalid ASN.1 syntax at line 1, column 35: "
                         "'A DEFINITIONS ::= BEGIN a INTEGER >!<END': "
                         "Expected \"::=\".")

    def test_parse_error_sequence_missing_type(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN'
                                   '  A ::= SEQUENCE { a } '
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 45: 'A DEFINITIONS ::= BEGIN  "
            "A ::= SEQUENCE { a >!<} END': Expected {CHOICE | INTEGER | REAL | "
            "BIT STRING | OCTET STRING | ENUMERATED | SEQUENCE OF | SEQUENCE | "
            "SET OF | SET | OBJECT IDENTIFIER | BOOLEAN | ANY DEFINED BY | ...}.")

    def test_parse_error_sequence_missing_member_name(self):
        with self.assertRaises(asn1tools.ParseError) as cm:
            asn1tools.parse_string('A DEFINITIONS ::= BEGIN'
                                   '  A ::= SEQUENCE { A } '
                                   'END')

        self.assertEqual(
            str(cm.exception),
            "Invalid ASN.1 syntax at line 1, column 43: 'A DEFINITIONS ::= "
            "BEGIN  A ::= SEQUENCE { >!<A } END': Expected \"}\".")


if __name__ == '__main__':
    unittest.main()



import unittest
import asn1tools


class Asn1ToolsDerTest(unittest.TestCase):

    maxDiff = None

    def test_encode_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn', 'der')

        self.assertEqual(all_types.encode('Boolean', True), b'\x01\x01\x01')
        self.assertEqual(all_types.encode('Integer', 32768), b'\x02\x03\x00\x80\x00')
        self.assertEqual(all_types.encode('Integer', 32767), b'\x02\x02\x7f\xff')
        self.assertEqual(all_types.encode('Integer', 256), b'\x02\x02\x01\x00')
        self.assertEqual(all_types.encode('Integer', 255), b'\x02\x02\x00\xff')
        self.assertEqual(all_types.encode('Integer', 128), b'\x02\x02\x00\x80')
        self.assertEqual(all_types.encode('Integer', 127), b'\x02\x01\x7f')
        self.assertEqual(all_types.encode('Integer', 1), b'\x02\x01\x01')
        self.assertEqual(all_types.encode('Integer', 0), b'\x02\x01\x00')
        self.assertEqual(all_types.encode('Integer', -1), b'\x02\x01\xff')
        self.assertEqual(all_types.encode('Integer', -128), b'\x02\x01\x80')
        self.assertEqual(all_types.encode('Integer', -129), b'\x02\x02\xff\x7f')
        self.assertEqual(all_types.encode('Integer', -256), b'\x02\x02\xff\x00')
        self.assertEqual(all_types.encode('Integer', -32768), b'\x02\x02\x80\x00')
        self.assertEqual(all_types.encode('Integer', -32769), b'\x02\x03\xff\x7f\xff')
        self.assertEqual(all_types.encode('Bitstring', (b'\x80', 1)), b'\x03\x02\x07\x80')
        self.assertEqual(all_types.encode('Octetstring', b'\x00'), b'\x04\x01\x00')
        self.assertEqual(all_types.encode('Null', None), b'\x05\x00')
        self.assertEqual(all_types.encode('Objectidentifier', '1.2'), b'\x06\x01\x2a')
        self.assertEqual(all_types.encode('Enumerated', 'one'), b'\x0a\x01\x01')
        self.assertEqual(all_types.encode('Utf8string', 'foo'), b'\x0c\x03foo')
        self.assertEqual(all_types.encode('Sequence', {}), b'\x30\x00')
        self.assertEqual(all_types.encode('Set', {}), b'\x31\x00')
        self.assertEqual(all_types.encode('Sequence2', {'a': 1}), b'\x30\x03\x02\x01\x01')
        self.assertEqual(all_types.encode('Set2', {'a': 2}), b'\x31\x03\x02\x01\x02')
        self.assertEqual(all_types.encode('Numericstring', '123'), b'\x12\x03123')
        self.assertEqual(all_types.encode('Printablestring', 'foo'), b'\x13\x03foo')
        self.assertEqual(all_types.encode('Ia5string', 'bar'), b'\x16\x03bar')
        self.assertEqual(all_types.encode('Universalstring', 'bar'), b'\x1c\x03bar')
        self.assertEqual(all_types.encode('Visiblestring', 'bar'), b'\x1a\x03bar')
        self.assertEqual(all_types.encode('Bmpstring', b'bar'), b'\x1e\x03bar')
        self.assertEqual(all_types.encode('Teletexstring', b'fum'), b'\x14\x03fum')
        self.assertEqual(all_types.encode('Utctime', '010203040506'),
                         b'\x17\x0d010203040506Z')
        self.assertEqual(all_types.encode('SequenceOf', []), b'0\x00')
        self.assertEqual(all_types.encode('SetOf', []), b'1\x00')

    def test_decode_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn', 'der')

        self.assertEqual(all_types.decode('Boolean', b'\x01\x01\x01'), True)
        self.assertEqual(all_types.decode('Integer', b'\x02\x01\x01'), 1)
        self.assertEqual(all_types.decode('Bitstring', b'\x03\x02\x07\x80'), (b'\x80', 1))
        self.assertEqual(all_types.decode('Octetstring', b'\x04\x01\x00'), b'\x00')
        self.assertEqual(all_types.decode('Null', b'\x05\x00'), None)
        self.assertEqual(all_types.decode('Objectidentifier', b'\x06\x01\x2a'), '1.2')
        self.assertEqual(all_types.decode('Enumerated', b'\x0a\x01\x01'), 'one')
        self.assertEqual(all_types.decode('Utf8string', b'\x0c\x03foo'), 'foo')
        self.assertEqual(all_types.decode('Sequence', b'\x30\x00'), {})
        self.assertEqual(all_types.decode('Set', b'\x31\x00'), {})
        self.assertEqual(all_types.decode('Sequence2', b'\x30\x00'), {'a': 0})
        self.assertEqual(all_types.decode('Sequence2', b'\x30\x03\x02\x01\x01'), {'a': 1})
        self.assertEqual(all_types.decode('Set2', b'\x31\x00'), {'a': 1})
        self.assertEqual(all_types.decode('Set2', b'\x31\x03\x02\x01\x02'), {'a': 2})
        self.assertEqual(all_types.decode('Numericstring', b'\x12\x03123'), '123')
        self.assertEqual(all_types.decode('Printablestring', b'\x13\x03foo'), 'foo')
        self.assertEqual(all_types.decode('Ia5string', b'\x16\x03bar'), 'bar')
        self.assertEqual(all_types.decode('Universalstring', b'\x1c\x03bar'), 'bar')
        self.assertEqual(all_types.decode('Visiblestring', b'\x1a\x03bar'), 'bar')
        self.assertEqual(all_types.decode('Bmpstring', b'\x1e\x03bar'), b'bar')
        self.assertEqual(all_types.decode('Teletexstring', b'\x14\x03fum'), b'fum')
        self.assertEqual(all_types.decode('Utctime', b'\x17\x0d010203040506Z'),
                         '010203040506')
        self.assertEqual(all_types.decode('SequenceOf', b'0\x00'), [])
        self.assertEqual(all_types.decode('SetOf', b'1\x00'), [])

    def test_repr_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'der')

        self.assertEqual(repr(all_types.types['Boolean']), 'Boolean(Boolean)')
        self.assertEqual(repr(all_types.types['Integer']), 'Integer(Integer)')
        self.assertEqual(repr(all_types.types['Bitstring']), 'BitString(Bitstring)')
        self.assertEqual(repr(all_types.types['Octetstring']), 'OctetString(Octetstring)')
        self.assertEqual(repr(all_types.types['Null']), 'Null(Null)')
        self.assertEqual(repr(all_types.types['Objectidentifier']),
                         'ObjectIdentifier(Objectidentifier)')
        self.assertEqual(repr(all_types.types['Enumerated']), 'Enumerated(Enumerated)')
        self.assertEqual(repr(all_types.types['Utf8string']), 'UTF8String(Utf8string)')
        self.assertEqual(repr(all_types.types['Sequence']), 'Sequence(Sequence, [])')
        self.assertEqual(repr(all_types.types['Set']), 'Set(Set, [])')
        self.assertEqual(repr(all_types.types['Sequence2']),
                         'Sequence(Sequence2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Set2']), 'Set(Set2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Numericstring']),
                         'NumericString(Numericstring)')
        self.assertEqual(repr(all_types.types['Printablestring']),
                         'PrintableString(Printablestring)')
        self.assertEqual(repr(all_types.types['Ia5string']), 'IA5String(Ia5string)')
        self.assertEqual(repr(all_types.types['Universalstring']),
                         'UniversalString(Universalstring)')
        self.assertEqual(repr(all_types.types['Visiblestring']),
                         'VisibleString(Visiblestring)')
        self.assertEqual(repr(all_types.types['Bmpstring']),
                         'BMPString(Bmpstring)')
        self.assertEqual(repr(all_types.types['Teletexstring']),
                         'TeletexString(Teletexstring)')
        self.assertEqual(repr(all_types.types['Utctime']), 'UTCTime(Utctime)')
        self.assertEqual(repr(all_types.types['SequenceOf']),
                         'SequenceOf(SequenceOf, Integer())')
        self.assertEqual(repr(all_types.types['SetOf']), 'SetOf(SetOf, Integer())')


if __name__ == '__main__':
    unittest.main()



import sys
import unittest

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import asn1tools


class Asn1ToolsCommandLineTest(unittest.TestCase):

    maxDiff = None

    def test_command_line_decode_ber_foo_question(self):
        argv = ['asn1tools',
                'decode',
                'tests/files/foo.asn',
                'Question',
                '300e0201011609497320312b313d333f']

        stdout = sys.stdout
        sys.argv = argv
        sys.stdout = StringIO()

        try:
            asn1tools._main()
        finally:
            actual_output = sys.stdout.getvalue()
            sys.stdout = stdout

        self.assertTrue('id: 1' in actual_output)
        self.assertTrue('question: Is 1+1=3?' in actual_output)

    def test_command_line_decode_uper_foo_question(self):
        argv = ['asn1tools',
                'decode',
                '--codec', 'uper',
                'tests/files/foo.asn',
                'Question',
                '01010993cd03156c5eb37e']

        stdout = sys.stdout
        sys.argv = argv
        sys.stdout = StringIO()

        try:
            asn1tools._main()
        finally:
            actual_output = sys.stdout.getvalue()
            sys.stdout = stdout

        self.assertTrue('id: 1' in actual_output)
        self.assertTrue('question: Is 1+1=3?' in actual_output)

    def test_command_line_decode_ber_rrc_bcch_dl_sch_message(self):
        argv = ['asn1tools',
                'decode',
                'tests/files/snmp_v1.asn',
                'Message',
                '30819f02010004067075626c6963a3819102013c020100020100308185302206'
                '122b06010401817d08330a0201070a86deb735040c3137322e33312e31392e37'
                '33301706122b06010401817d08330a0201050a86deb960020102302306122b06'
                '010401817d08330a0201070a86deb736040d3235352e3235352e3235352e3030'
                '2106122b06010401817d08330a0201070a86deb738040b3137322e33312e3139'
                '2e32']

        stdout = sys.stdout
        sys.argv = argv
        sys.stdout = StringIO()

        try:
            asn1tools._main()
        finally:
            actual_output = sys.stdout.getvalue()
            sys.stdout = stdout

        expected_output = [
            "community: '7075626c6963'",
            "data:",
            "  set-request:",
            "    variable-bindings:",
            "      [0]:",
            "        value:",
            "          simple:",
            "            string: '3137322e33312e31392e3733'",
            "        name: 1.3.6.1.4.1.253.8.51.10.2.1.7.10.14130101",
            "      [1]:",
            "        value:",
            "          simple:",
            "            number: 2",
            "        name: 1.3.6.1.4.1.253.8.51.10.2.1.5.10.14130400",
            "      [2]:",
            "        value:",
            "          simple:",
            "            string: '3235352e3235352e3235352e30'",
            "        name: 1.3.6.1.4.1.253.8.51.10.2.1.7.10.14130102",
            "      [3]:",
            "        value:",
            "          simple:",
            "            string: '3137322e33312e31392e32'",
            "        name: 1.3.6.1.4.1.253.8.51.10.2.1.7.10.14130104",
            "    error-index: 0",
            "    error-status: 0",
            "    request-id: 60",
            "version: 0"
        ]

        for line in expected_output:
            self.assertTrue(line in actual_output)


if __name__ == '__main__':
    unittest.main()
